#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <pthread.h>

#include <sys/ioctl.h>
#include <sys/mman.h>

#include "kvm.h"

#define MEM_SIZE	(2ULL * 1024 * 1024 * 1024)
#define VCPU_NUM	6
#define MAX_VCPU	512

static char *readonly_buffer;
static unsigned int readonly_pages;

static struct {
	char p[5 * PAGE_SIZE];
} stack[MAX_VCPU];

#define PRINT_PORT 0x9
#define EXIT_PORT	0xf
#define FAULT_PORT	0xe

static struct {
	char p[PAGE_SIZE];
} print_buf[MAX_VCPU];

extern char _end[];
extern char _start[];

static void vm_init_memory_layout(struct vm *vm, int vcpu_num, u64 write_mem_size)
{
	u64 programe_size = (long)_end;

	/*
	 * Mapping the programe to guest memory., so in guest, we can call the functions in this programe.
	 */
	programe_size = PAGE_ALIGN(programe_size);

	readonly_pages = write_mem_size / PAGE_SIZE;
	readonly_buffer = wmalloc_align(readonly_pages * PAGE_SIZE);
		
	kvm_register_mem_readonly(vm, (u64)(u32)readonly_buffer, readonly_buffer,  readonly_pages * PAGE_SIZE);
	kvm_register_mem(vm, 0, 0, programe_size);
}

#ifdef DEBUG
#define vprint(vcpu_id, fmt, args...)	do {	\
	wsprintf(print_buf[vcpu_id].p, fmt, ##args);	\
	asm volatile("outb %%al, %%dx" : : "a"(0x9), "d"(0x9));} while (0)
#else
#define vprint(vcpu_id, fmt, args...)
#endif

#define vprints(vcpu_id, s)	do {	\
	our_memcpy(print_buf[vcpu_id].p, s, PAGE_SIZE);	\
	asm volatile("outb %%al, %%dx" : : "a"(0x9), "d"(0x9));} while (0)

static int vcpu_readonly_test(int long vcpu_id)
{
	char c;
	unsigned int i = 0;

	vprints(vcpu_id, "Readonly Read------.\n");
	for (i = 0; i < readonly_pages; i++)
		c = readonly_buffer[i * PAGE_SIZE];

	vprints(vcpu_id, "Readonly test: Write------.\n");

	for (i = 0; i < readonly_pages; i++) {
		readonly_buffer[i * PAGE_SIZE] = 'K';

		/* Avoid the compiler optimizing the write and read. */
		mb();

		c = readonly_buffer[i * PAGE_SIZE];
		if (c != 'K')
			return -1;
	}

	return 0;
}

static void vcpu_common_fun(void)
{
	int long vcpu_id;

	asm ("mov %%esi, %0" : "=rm"(vcpu_id));

	vprints(vcpu_id, "UP.\n");

	vprint(vcpu_id, "vcpu:%d up.\n", vcpu_id);

	if (vcpu_readonly_test(vcpu_id))
		asm volatile("outb %%al, %%dx" : : "a"(FAULT_PORT), "d"(FAULT_PORT));

	asm volatile("outb %%al, %%dx" : : "a"(EXIT_PORT), "d"(EXIT_PORT));
}

static struct vm *vm;

static int kvm_io(struct vcpu *vcpu, u16 port, int direction)
{
	if (port != PRINT_PORT || direction != 1)
		return -1;

	(void)vcpu;
	dprintf("VCPU[%d]: %s", vcpu->id, print_buf[vcpu->id].p);
	return 0;
}

static int kvm_mmio(u64 gpa, int len, u8 *data, int write, int readonly_fault)
{

	dprintf("MMIO: gpa:%llx, len:%x, value:%c, write:%d, readonly-fault:%d.\n", gpa, len, data[0], write, readonly_fault);

	if (gpa < (unsigned long)readonly_buffer ||gpa > (unsigned long)(readonly_buffer + readonly_pages * PAGE_SIZE))
		return -1;

	if (!write || len != 1)
		return -1;

	if (!readonly_fault)
		return -1;

	memcpy((void *)(unsigned long)gpa, data, len);
	return 0;
}

static void vcpu_init_stack(struct vcpu *vcpu)
{
	vcpu->stack = (void *)PAGE_ALIGN((long)stack[vcpu->id].p);

	dprintf("vcpu:%d stack:%p.\n", vcpu->id, vcpu->stack);
}

static void *vcpu_thread (void *arg)
{
	unsigned long vcpu_id = (unsigned long)arg;
	struct vcpu *vcpu;

	vcpu = create_vcpu(vm, vcpu_id);
	vcpu_init_stack(vcpu);
	vcpu_register_fun(vcpu, vcpu_common_fun);

run:
	vcpu_run(vcpu);

	dprintf("Exit code:%x.\n", vcpu->kvm_run->exit_reason);
	switch (vcpu->kvm_run->exit_reason) {
		case KVM_EXIT_IO: {
			if (vcpu->kvm_run->io.port == EXIT_PORT)
				return (void *)0;

			if (!kvm_io(vcpu, vcpu->kvm_run->io.port, vcpu->kvm_run->io.direction))
					goto run;
			break;
		}

		case KVM_EXIT_MMIO: {
			if (!kvm_mmio(vcpu->kvm_run->mmio.phys_addr, vcpu->kvm_run->mmio.len, vcpu->kvm_run->mmio.data, vcpu->kvm_run->mmio.is_write,
				vcpu->kvm_run->mmio.write_readonly_mem))
				goto run;
			
			break;
		}

		default:
			break;
	}			

	return (void *)-1;
}

#define log_size(s) ((s >> 12) + 7) / 8

static void usage(void)
{
	printf("Usage:\n");
	printf("-c:	the vcpu number (default: %d).\n", VCPU_NUM);
	printf("-m:	the memory size (M) (default: %lld M).\n", MEM_SIZE / 1024 /1024);
	printf("-t:	the memory write time (default: %d M).\n", 10);
}

int main(int argc, char *argv[])
{
	u64 ram_size;
	long vcpu_num, count, i, opt;
	pthread_t *tinfo;
	char *endptr;
	int r;

	(void)argc; (void)argv;

	dprintf("Start...\n");

	ram_size = MEM_SIZE;
	vcpu_num = VCPU_NUM;
	count = 10;

	while ((opt = getopt(argc, argv, "c:m:t:")) != -1) {
		switch (opt) {
		case 'c':
			vcpu_num = strtol(optarg, &endptr, 10);
			break;
		case 'm':
			ram_size = strtol(optarg, &endptr, 10);
			ram_size *= 1024 * 1024;
			break;
		case 't':
			count = strtol(optarg, &endptr, 10);
			break;
		default:
			usage();
			fmt_die("Unknown %c.\n", opt);
		}
	}

	ram_size = PAGE_ALIGN(ram_size);

	printf("VCPU:%ld, Mem:%lldM, Time:%ld.\n", vcpu_num, ram_size / 1024 / 1024, count);

	tinfo = wmalloc(sizeof(pthread_t) * vcpu_num);

	vm = create_vm();

	r = kvm_check_extension(vm, KVM_CAP_READONLY_MEM);
	if (r != 1)
		die("READONLY MEM is not supported.\n");

	vm_init_memory_layout(vm, vcpu_num, ram_size);

	for (i = 0; i < vcpu_num; i++)
		pthread_create(&tinfo[i], NULL, vcpu_thread, (void *)(unsigned long)i);

	for (i = 0; i < vcpu_num; i++) {
		void *ret;

		if (pthread_join(tinfo[i], &ret) < 0)
			perror("pthread join");

		dprintf("thread %ld ret:%d.\n", i, (int)ret);
		if ((int)ret) {
			printf("Readonly test fail.\n");
			return (int)ret;
		}
	}

	printf("Readonly test okay.\n");
	return 0;
}
